/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of libSVDlp: a SVD calculation library
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "libSVDlp.h"
	using std::cout;
	using std::endl;
#ifdef USE_EIGEN
	using Eigen::MatrixXd;

void prepareMatrix( MatrixXd*& M, int nRows, int nCols )
{
	if ( M ) {
		eigen_assert( ( M->rows() == nRows ) && ( M->cols() == nCols ) );
	} else {
		M = new MatrixXd( nRows, nCols );
	}
}

void prepareMatrix( MatrixXd*& M, int nRows, int nCols, integer& ld )
{
	prepareMatrix( M, nRows, nCols );
	ld = nRows;
}

int svdWrapper( MatrixXd& A, SVD_TYPE svdType, MatrixXd*& S, MatrixXd*& U, MatrixXd*& VT )
{
	integer m = A.rows();
	integer n = A.cols();

	integer ldA  = m;
	integer ldU  = 1;
	integer ldVT = 1;

	const int minMN = min( m, n );
	const int maxMN = max( m, n );

	integer info = 0;

	integer workSize = -1;	

	double optWorkSize = -1;

	double* work = NULL;	

	if ( svdType == SVD_COMPLETE ) {
		prepareMatrix( S,  1, minMN );
		prepareMatrix( U,  m, m, ldU );
		prepareMatrix( VT, n, n, ldVT );			
		dgesvd_( "A", "A", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, &optWorkSize, &workSize, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesvd_( "A", "A", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, work, &workSize, &info );
	} else if ( svdType == SVD_VALUES ) {
		prepareMatrix( S,  1, minMN );
		dgesvd_( "N", "N", &m, &n, A.data(), &ldA, S->data(), NULL, &ldU, NULL, &ldVT, &optWorkSize, &workSize, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesvd_( "N", "N", &m, &n, A.data(), &ldA, S->data(), NULL, &ldU, NULL, &ldVT, work, &workSize, &info );
	} else if ( svdType == SVD_SUBSET ) {
		ldVT = minMN;
		prepareMatrix( S,  1, minMN );
		prepareMatrix( U,  m, minMN, ldU );
		prepareMatrix( VT, minMN, n, ldVT );
		dgesvd_( "S", "S", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, &optWorkSize, &workSize, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesvd_( "S", "S", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, work, &workSize, &info );
	} else {
		cout << "unkown SVD type" << endl;
	}

	free( work );

	return 0;
}

int sddWrapper( MatrixXd& A, SVD_TYPE svdType, MatrixXd*& S, MatrixXd*& U, MatrixXd*& VT )
{
	integer m = A.rows();
	integer n = A.cols();

	integer ldA  = m;
	integer ldU  = 1;
	integer ldVT = 1;

	const int minMN = min( m, n );
	const int maxMN = max( m, n );

	integer workSize = -1;

	double optWorkSize = -1;

	double* work = NULL;

	integer* iWork = ( integer* )malloc( 8ul * minMN * sizeof( integer ) );

	integer info = 0;

	if ( svdType == SVD_COMPLETE ) {
		prepareMatrix( S,  1, minMN );
		prepareMatrix( U,  m, m, ldU );
		prepareMatrix( VT, n, n, ldVT );
		dgesdd_( "A", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, &optWorkSize, &workSize, iWork, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesdd_( "A", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, work, &workSize, iWork, &info );
	} else if ( svdType == SVD_VALUES ) {
		prepareMatrix( S,  1, minMN );
		dgesdd_( "N", &m, &n, A.data(), &ldA, S->data(), NULL, &ldU, NULL, &ldVT, &optWorkSize, &workSize, iWork, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesdd_( "N", &m, &n, A.data(), &ldA, S->data(), NULL, &ldU, NULL, &ldVT, work, &workSize, iWork, &info );
	} else if ( svdType == SVD_SUBSET ) {
		ldVT = minMN;
		prepareMatrix( S,  1, minMN );
		prepareMatrix( U,  m, minMN, ldU );
		prepareMatrix( VT, minMN, n, ldVT );
		dgesdd_( "S", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, &optWorkSize, &workSize, iWork, &info );
		workSize = static_cast<integer>( optWorkSize );
		work = ( double* )malloc( workSize * sizeof( double ) );
		dgesdd_( "S", &m, &n, A.data(), &ldA, S->data(), U->data(), &ldU, VT->data(), &ldVT, work, &workSize, iWork, &info );
	} else {
		cout << "unkown SVD type" << endl;
	}

	free( work );
	free( iWork );

	return 0;
}

#endif
